#include "reader.h"

volatile sig_atomic_t	x_stop_server;
int			x_nlinks;
pid_t			x_child[DX_MAX_LINKS];
char			x_home_dir[256];

static struct option x_opts[] = {
  { "ipaddr", 1, NULL, 'a' },
  { "port",   1, NULL, 'p' },
  { "config", 1, NULL, 'c' },
  {  NULL,    0, NULL,  0  }
};

static void dx_show_usage(const char *pr_name)
{
    fprintf(stderr,
        "usage: %s [options] start|stop\n"
        "    -a, --ipaddr   IP address\n"
        "    -p, --port     port\n"
        "    -c, --config   config file\n"
        "    start          start reader\n"
        "    stop           stop reader\n"
        "\n"
        , pr_name
    );
}

static int dx_stop_server(void)
{
    int		cnt;
    pid_t	pid;
    FILE	*fp;
    char	filename[256];

    sprintf(filename, "%s/xreader.pid", x_home_dir);
    if ((fp = fopen(filename, "r")) == NULL) {
        fprintf(stderr, "open pid file \"%s\" error:\n%s\n", filename, strerror(errno));
        return (-1);
    }

    fscanf(fp, "%d", &pid);
    fclose(fp);

    cnt = 3;
    while (cnt > 0) {
        kill(pid, SIGTERM);
        sleep(1);
        if (kill(pid, 0) < 0)
            break;
        cnt--;
    }

    if (cnt > 0) {
        fprintf(stderr, "xreader server terminated !\n\n");
        return (0);
    }

    kill(pid, SIGKILL);
    fprintf(stderr, "xreader server killed !!!\n\n");

    return (0);
}

static int dx_daemonize(void)
{
    int		fd;
    pid_t	pid;

    if ((pid=fork()) == -1)
        return (-1);
    else if (pid != 0)
        exit(0);

    setsid();

    if ((pid=fork()) == -1)
        return (-1);
    else if (pid > 0)
        exit(0);

    close(fileno(stdin));
    close(fileno(stdout));
    close(fileno(stderr));

    fd = open("/dev/null", O_RDWR);	/* stdin */
    dup(fd);	/* stdout */
    dup(fd);	/* stderr */

    umask(0);

    return (0);
}

int main(int argc, char *argv[])
{
    int		i;
    int		cmd;
    int		listen_sd;
    int		new_sd;
    int		optc;
    int		opt_idx;
    int		port;
    pid_t	pid;
    FILE	*fp = NULL;
    char	*env;
    const char	*cmdopts = "a:p:c:";
    char	ip[16];
    char	filename[256];


    /* parse arguments */
    while ((optc =
            getopt_long(argc, argv, cmdopts, x_opts, &opt_idx)
           ) != -1) {
        switch (optc) {
        case 'a':
            strncpy(ip, optarg, 16);
            ip[15] = '\0';
            break;
        case 'p':
            port = atoi(optarg);
            break;
        case 'c':
            strncpy(filename, optarg, 256);
            filename[255] = '\0';
            break;
        default:
            dx_show_usage(argv[0]);
            return (-1);
        }
    }

    if (optind >= argc) {
        dx_show_usage(argv[0]);
        return (-1);
    }

    if (!strcmp(argv[optind], "start")) {
        cmd = DX_SERVER_START;
    } else if (!strcmp(argv[optind], "stop")) {
        cmd = DX_SERVER_STOP;
    } else {
        dx_show_usage(argv[0]);
        return (-1);
    }

    if ((env = getenv("XREADER_HOME")) == NULL) {
        fprintf(stderr, "env \"XREADER_HOME\" not set.\n");
        return (-1);
    }

    strncpy(x_home_dir, env, 256);
    x_home_dir[255] = '\0';
    i = strlen(x_home_dir) - 1;
    if (x_home_dir[i] == '/')
        x_home_dir[i] = '\0';
    

    if (cmd == DX_SERVER_STOP) {
        dx_stop_server();
        return (0);
    }

    /* goto daemon mode */
    if (dx_daemonize() < 0) {
        fprintf(stderr, "go to DAEMON mode error.\n%s\n", strerror(errno));
        return (-1);
    }

    /* set signal */
    if (dx_setsignals() < 0) {
        dx_logfile_error("set signals error.");
        return (-1);
    }

    sprintf(filename, "%s/xreader.pid", x_home_dir);
    if ((fp = fopen(filename, "w")) == NULL) {
        dx_logfile_error("open pid file error: %s", strerror(errno));
        return (-1);
    }

    fprintf(fp, "%d", getpid());
    fclose(fp);

    /* create listener */
    if ((listen_sd = dx_create_listener(ip, port)) < 0) {
        dx_logfile_error("create listener error.");
        return (-1);
    }

    /* init child pid array */
    for (i=0; i<DX_MAX_LINKS; i++) {
        x_child[i] = -1;
    }

    x_nlinks = 0;
    /* loop for accept connection */
    while (x_stop_server == 0) {
        if ((new_sd = dx_accept_client(listen_sd)) < 0) {
            if (errno == EINTR)
                continue;
            dx_logfile_error("accept client connection error.");
            sleep(1);
            continue;
        }

        if (x_nlinks >= DX_MAX_LINKS) {
            dx_logfile_error("client links reach the limit: %d", DX_MAX_LINKS);
            close(new_sd);
            continue;
        }

        if ((pid = fork()) < 0) {
            dx_logfile_error("for(): %s", strerror(errno));
            x_stop_server = 1;
            continue;
        }

        if (pid == 0) {
            close(listen_sd);
            if (fcntl(new_sd, F_SETFL, O_NONBLOCK) < 0) {
                dx_logfile_error("fcntl(): %s", strerror(errno));
            }

            dx_worker(new_sd);
            close(new_sd);
            dx_close_logfile();
            exit(0);
        }

        /* set child pid */
        close(new_sd);
        for (i=0; i<DX_MAX_LINKS; i++) {
            if (x_child[i] == -1) {
                x_child[i] = pid;
                x_nlinks++;
                break;
            }
        }

        if (i >= DX_MAX_LINKS) {
            kill(pid, SIGTERM);
        }
    }

    shutdown(listen_sd, SHUT_RDWR);

    /* kill child first */
    for (i=0; i<DX_MAX_LINKS; i++) {
        if (x_child[i] == -1)
            continue;
        kill(x_child[i], SIGTERM);
        waitpid(x_child[i], NULL, WNOHANG);
        x_child[i] = -1;
    }

    close(listen_sd);
    dx_close_logfile();

    return (0);
}
